"""
Code for clustering of multivariate (2- or 3-dimensional) time-series data.
See the documentation for all the details.
"""

import copy
import os
import shutil
from typing import List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.image import NonUniformImage
from matplotlib.patches import Ellipse
from scipy.stats import gaussian_kde

from onion_clustering.classes import ClusteringObject2D
from onion_clustering.first_classes import MultiData, Parameters, StateMulti
from onion_clustering.functions import (
    custom_fit,
    find_half_height_around_max,
    find_minima_around_max,
    max_prob_assignment_2d,
    param_grid,
    read_input_data,
    relabel_states_2d,
)

OUTPUT_FILE = "states_output.txt"


def all_the_input_stuff(number_of_sigmas: float) -> ClusteringObject2D:
    """
    Reads input parameters and raw data from specified files and directories,
    processes the raw data, and creates output files.
    """
    # Read input parameters from files.
    data_directory = read_input_data()
    par = Parameters("input_parameters.txt")
    par.print_to_screen()

    data = MultiData(data_directory)
    data.remove_delay(par.t_delay)

    ### Create files for output
    with open(OUTPUT_FILE, "w", encoding="utf-8") as file:
        file.write("#")
    figures_folder = "output_figures"
    if not os.path.exists(figures_folder):
        os.makedirs(figures_folder)
    for filename in os.listdir(figures_folder):
        file_path = os.path.join(figures_folder, filename)
        try:
            if os.path.isfile(file_path):
                os.remove(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except OSError as exc_msg:
            print(f"Failed to delete {file_path}. Reason: {exc_msg}")

    clustering_object = ClusteringObject2D(par, data, number_of_sigmas)

    return clustering_object


def gauss_fit_max(
    m_clean: np.ndarray,
    m_limits: np.ndarray,
    bins: Union[int, str],
    number_of_sigmas: float,
    filename: str,
    full_out: bool,
) -> Union[StateMulti, None]:
    """
    Perform Gaussian fit and generate plots based on the provided data.

    Parameters
    ----------

    m_clean : np.ndarray of shape (n_particles, n_timesteps)
        Processed array of cleaned data.

    m_limits : list[list[int]]
        List containing minimum and maximum values for each dimension.

    bins : Union[int, str]
        Number of bins for histograms or 'auto' for automatic binning.

    number_of_sigmas : float
        To set the thresholds for assigning windows to the state.

    filename : str
        Name of the output file to save the plot.

    full_out : bool
        If True, plot all the intermediate histograms with the best fit.
        Useful for debugging.

    Returns
    -------

    State : StateMulti
        Object containing Gaussian fit parameters (mu, sigma, area)
        or None if the fit fails.
    """
    print("* Gaussian fit...")

    ### 1. Create the histogram with KDE estimator ###
    if m_clean.ndim == 3:
        flat_m = m_clean.reshape(
            m_clean.shape[0] * m_clean.shape[1], m_clean.shape[2]
        ).T
    elif m_clean.ndim == 2:
        flat_m = m_clean.T

    lims = np.zeros((flat_m.shape[0], 2))
    for i, data in enumerate(flat_m):
        lims[i][0] = np.min(data)
        lims[i][1] = np.max(data)

    try:
        if flat_m.shape[0] == 2:
            X, Y = np.mgrid[
                lims[0][0] : lims[0][1] : 100j, lims[1][0] : lims[1][1] : 100j
            ]
            edges = np.array([X[:, 0], Y[0, :]])
            positions = np.vstack([X.ravel(), Y.ravel()])
            values = np.vstack([flat_m[0], flat_m[1]])
            kernel = gaussian_kde(values)
            counts = np.reshape(kernel(positions).T, X.shape)
        elif flat_m.shape[0] == 3:
            X, Y, Z = np.mgrid[
                lims[0][0] : lims[0][1] : 100j,
                lims[1][0] : lims[1][1] : 100j,
                lims[2][0] : lims[2][1] : 100j,
            ]
            edges = np.array([X[:, :, 0], Y[:, 0, :], Z[0, :, :]])
            positions = np.vstack([X.ravel(), Y.ravel(), Z.ravel()])
            values = np.vstack([flat_m[0], flat_m[1], flat_m[2]])
            kernel = gaussian_kde(values)
            counts = np.reshape(kernel(positions).T, X.shape)
    except np.linalg.LinAlgError:
        print("\tWARNING: not enough data for KDE.")
        return None

    ### 2. Find the maximum ###
    def find_max_index(data: np.ndarray):
        max_val = data.max()
        max_indices = np.argwhere(data == max_val)
        return max_indices[0]

    max_ind = find_max_index(counts)

    ### 3. Find the minima surrounding it ###
    gap = 3
    minima = find_minima_around_max(counts, max_ind, gap)

    ### 4. Try the fit between the minima and check its quality ###
    popt_min: List[float] = []
    det_coeff_min = 1.0
    flag_min = 1
    for dim in range(flat_m.shape[0]):
        tmp_flag_min, r2_min, popt = custom_fit(
            dim, max_ind[dim], minima, edges[dim], counts, lims
        )
        popt[2] *= flat_m[0].size
        popt_min.extend(popt)
        det_coeff_min *= r2_min
        flag_min *= tmp_flag_min

    ### 5. Find the interval of half height ###
    minima = find_half_height_around_max(counts, max_ind, gap)

    ### 6. Try the fit between the minima and check its quality ###
    popt_half: List[float] = []
    det_coeff_half = 1.0
    flag_half = 1
    for dim in range(flat_m.shape[0]):
        tmp_flag_half, r2_half, popt = custom_fit(
            dim, max_ind[dim], minima, edges[dim], counts, lims
        )
        popt[2] *= flat_m[0].size
        popt_half.extend(popt)
        det_coeff_half *= r2_half
        flag_half *= tmp_flag_half

    ### 7. Choose the best fit ###
    r2 = det_coeff_min
    if flag_min == 1 and flag_half == 0:
        popt = np.array(popt_min)
    elif flag_min == 0 and flag_half == 1:
        popt = np.array(popt_half)
        r2 = det_coeff_half
    elif flag_min * flag_half == 1:
        if det_coeff_min >= det_coeff_half:
            popt = np.array(popt_min)
        else:
            popt = np.array(popt_half)
            r2 = det_coeff_half
    else:
        print("\tWARNING: this fit is not converging.")
        return None

    if len(popt) != flat_m.shape[0] * 3:
        print("\tWARNING: this fit is not converging.")
        return None

    ### Find the tresholds for state identification
    mean, sigma, area = [], [], []
    for dim in range(flat_m.shape[0]):
        mean.append(popt[3 * dim])
        sigma.append(popt[3 * dim + 1])
        area.append(popt[3 * dim + 2])
    state = StateMulti(np.array(mean), np.array(sigma), np.array(area))
    state.build_boundaries(number_of_sigmas)

    ### Plot the distribution and the fitted Gaussians
    if flat_m.shape[0] == 2:
        with open(OUTPUT_FILE, "a", encoding="utf-8") as file:
            print("\n", file=file)
            print(
                f"\tmu = [{popt[0]:.4f}, {popt[3]:.4f}],"
                f" sigma = [{popt[1]:.4f}, {popt[4]:.4f}]"
            )
            print(
                f"\tmu = [{popt[0]:.4f}, {popt[3]:.4f}],"
                f" sigma = [{popt[1]:.4f}, {popt[4]:.4f}],"
                f" area = {popt[2]:.4f}, {popt[5]:.4f}",
                file=file,
            )
            print(f"\tFit r2 = {r2}", file=file)

        if full_out:
            fig, ax = plt.subplots(figsize=(6, 6))
            img = NonUniformImage(ax, interpolation="nearest")
            xcenters = (edges[0] + edges[0]) / 2
            ycenters = (edges[1] + edges[1]) / 2
            img.set_data(xcenters, ycenters, counts.T)
            ax.add_image(img)
            ax.scatter(mean[0], mean[1], s=8.0, c="red")
            circle1 = Ellipse(
                tuple(mean), sigma[0], sigma[1], color="r", fill=False
            )
            circle2 = Ellipse(
                tuple(mean),
                state.axis[0],
                state.axis[1],
                color="r",
                fill=False,
            )
            ax.add_patch(circle1)
            ax.add_patch(circle2)
            ax.set_xlim(m_limits[0][0], m_limits[0][1])
            ax.set_ylim(m_limits[1][0], m_limits[1][1])

            fig.savefig(filename + ".png", dpi=600)
            plt.close(fig)

    elif flat_m.shape[0] == 3:
        with open(OUTPUT_FILE, "a", encoding="utf-8") as file:
            print("\n", file=file)
            print(
                f"\tmu = [{popt[0]:.4f}, {popt[3]:.4f}, {popt[6]:.4f}], "
                f"sigma = [{popt[1]:.4f}, {popt[4]:.4f}, {popt[7]:.4f}], "
                f"area = {popt[2]:.4f}, {popt[5]:.4f}, {popt[8]:.4f}"
            )
            print(
                f"\tmu = [{popt[0]:.4f}, {popt[3]:.4f}, {popt[6]:.4f}], "
                f"sigma = [{popt[1]:.4f}, {popt[4]:.4f}, {popt[7]:.4f}], "
                f"area = {popt[2]:.4f}, {popt[5]:.4f}, {popt[8]:.4f}",
                file=file,
            )
            print(f"\tFit r2 = {r2}", file=file)

        if full_out:
            fig, ax = plt.subplots(2, 2, figsize=(6, 6))
            xcenters = (edges[0] + edges[0]) / 2
            ycenters = (edges[1] + edges[1]) / 2
            zcenters = (edges[2] + edges[2]) / 2

            img = NonUniformImage(ax[0][0], interpolation="nearest")
            img.set_data(xcenters, ycenters, np.sum(counts, axis=0))
            ax[0][0].add_image(img)
            ax[0][0].scatter(mean[0], mean[1], s=8.0, c="red")
            circle1 = Ellipse(
                tuple([mean[0], mean[1]]),
                sigma[0],
                sigma[1],
                color="r",
                fill=False,
            )
            circle2 = Ellipse(
                tuple([mean[0], mean[1]]),
                state.axis[0],
                state.axis[1],
                color="r",
                fill=False,
            )
            ax[0][0].add_patch(circle1)
            ax[0][0].add_patch(circle2)

            img = NonUniformImage(ax[0][1], interpolation="nearest")
            img.set_data(zcenters, ycenters, np.sum(counts, axis=1))
            ax[0][1].add_image(img)
            ax[0][1].scatter(mean[2], mean[1], s=8.0, c="red")
            circle1 = Ellipse(
                tuple([mean[2], mean[1]]),
                sigma[2],
                sigma[1],
                color="r",
                fill=False,
            )
            circle2 = Ellipse(
                tuple([mean[2], mean[1]]),
                state.axis[2],
                state.axis[1],
                color="r",
                fill=False,
            )
            ax[0][1].add_patch(circle1)
            ax[0][1].add_patch(circle2)

            img = NonUniformImage(ax[1][0], interpolation="nearest")
            img.set_data(xcenters, zcenters, np.sum(counts, axis=2))
            ax[1][0].add_image(img)
            ax[1][0].scatter(mean[0], mean[2], s=8.0, c="red")
            circle1 = Ellipse(
                tuple([mean[0], mean[2]]),
                sigma[0],
                sigma[2],
                color="r",
                fill=False,
            )
            circle2 = Ellipse(
                tuple([mean[0], mean[2]]),
                state.axis[0],
                state.axis[2],
                color="r",
                fill=False,
            )
            ax[1][0].add_patch(circle1)
            ax[1][0].add_patch(circle2)

            ax[0][0].set_xlim(m_limits[0][0], m_limits[0][1])
            ax[0][0].set_ylim(m_limits[1][0], m_limits[1][1])
            ax[0][1].set_xlim(m_limits[2][0], m_limits[2][1])
            ax[0][1].set_ylim(m_limits[1][0], m_limits[1][1])
            ax[1][0].set_xlim(m_limits[0][0], m_limits[0][1])
            ax[1][0].set_ylim(m_limits[2][0], m_limits[2][1])

            fig.savefig(filename + ".png", dpi=600)
            plt.close(fig)

    return state


def find_stable_trj(
    m_clean: np.ndarray,
    tau_window: int,
    state: StateMulti,
    all_the_labels: np.ndarray,
    offset: int,
) -> Tuple[np.ndarray, float, bool]:
    """
    Find stable windows in the trajectory.

    Args:
    - m_clean (np.ndarray): Cleaned trajectory data.
    - tau_window (int): Length of the window for analysis.
    - state (StateMulti): State information.
    - all_the_labels (np.ndarray): All labels for the trajectory.
    - offset (int): Offset for labeling stable windows.

    Returns a Tuple of:
    - np.ndarray: Array of non-stable windows.
    - float: Fraction of stable windows.
    - bool: Indicates one last state after finding stable windows.
    """
    print("* Finding stable windows...")

    # Calculate the number of windows in the trajectory
    number_of_windows = int(m_clean.shape[1] / tau_window)

    mask_unclassified = all_the_labels < 0.5
    m_reshaped = m_clean[:, : number_of_windows * tau_window].reshape(
        m_clean.shape[0], number_of_windows, tau_window, m_clean.shape[2]
    )
    shifted = m_reshaped - state.mean
    rescaled = shifted / state.axis
    squared_distances = np.sum(rescaled**2, axis=3)
    mask_dist = np.max(squared_distances, axis=2) <= 1.0
    mask = mask_unclassified & mask_dist

    all_the_labels[mask] = offset + 1
    counter = np.sum(mask)

    # Store non-stable windows in a list, for the next iteration
    m_new = []
    mask_remaining = mask_unclassified & ~mask
    for i, window in np.argwhere(mask_remaining):
        r_w = m_clean[i, window * tau_window : (window + 1) * tau_window]
        m_new.append(r_w)

    # Calculate the fraction of stable windows found
    fraction_of_points = counter / (all_the_labels.size)

    # Print the fraction of stable windows
    with open(OUTPUT_FILE, "a", encoding="utf-8") as file:
        print(
            f"\tFraction of windows in state {offset + 1}"
            f" = {fraction_of_points:.3}"
        )
        print(
            f"\tFraction of windows in state {offset + 1}"
            f" = {fraction_of_points:.3}",
            file=file,
        )

    # Convert the list of non-stable windows to a NumPy array
    m_new_arr = np.array(m_new)
    if m_new_arr.size > 0:
        m_new_arr = m_new_arr.reshape(
            m_new_arr.shape[0] * m_new_arr.shape[1], m_new_arr.shape[2]
        )
    one_last_state = True
    if len(m_new_arr) == 0:
        one_last_state = False

    # Return the array of non-stable windows, the fraction of stable windows,
    # and the updated list_of_states
    return m_new_arr, fraction_of_points, one_last_state


def iterative_search(
    cl_ob: ClusteringObject2D,
    name: str,
    full_out: bool,
) -> Tuple[ClusteringObject2D, bool]:
    """
    Perform an iterative search to identify stable windows in trajectory data.

    Args:
    - name (str): Name for the output figures.

    Returns:
    - cl_ob (ClusteringObject): updated with the clustering results.
    - one_last_state (bool): Indicates if there's one last state remaining.
    """
    tau_w, bins = cl_ob.par.tau_w, cl_ob.par.bins

    # Initialize an array to store labels for each window.
    num_windows = int(cl_ob.data.num_of_steps / tau_w)
    tmp_labels = np.zeros((cl_ob.data.num_of_particles, num_windows)).astype(
        int
    )

    states_list = []
    m_copy = cl_ob.data.matrix
    iteration_id = 1
    states_counter = 0
    one_last_state = False
    while True:
        ### Locate and fit maximum in the signal distribution
        state = gauss_fit_max(
            m_copy,
            np.array(cl_ob.data.range),
            bins,
            cl_ob.number_of_sigmas,
            "output_figures/" + name + "Fig1_" + str(iteration_id),
            full_out,
        )
        if state is None:
            print("Iterations interrupted because fit does not converge. ")
            break

        ### Find the windows in which the trajectories are stable
        m_new, counter, one_last_state = find_stable_trj(
            cl_ob.data.matrix, tau_w, state, tmp_labels, states_counter
        )
        state.perc = counter

        if counter > 0.0:
            states_list.append(state)

        states_counter += 1
        iteration_id += 1
        ### Exit the loop if no new stable windows are found
        if counter <= 0.0:
            print("Iterations interrupted because last state is empty. ")
            break
        if m_new.size == 0:
            print("Iterations interrupted because all data points assigned. ")
            break
        m_copy = m_new

    cl_ob.iterations = len(states_list)
    all_the_labels, list_of_states = relabel_states_2d(tmp_labels, states_list)

    ############################################
    ### Here to insert probab_reassigmnet_2d ###
    all_the_labels, list_of_states = max_prob_assignment_2d(
        list_of_states,
        cl_ob.data.matrix,
        all_the_labels,
        tau_w,
        cl_ob.number_of_sigmas,
    )
    ############################################

    cl_ob.data.labels = all_the_labels
    cl_ob.states = list_of_states
    return cl_ob, one_last_state


def timeseries_analysis(
    cl_ob: ClusteringObject2D,
    tau_w: int,
    t_smooth: int,
    full_out: bool,
) -> Tuple[int, float, List[float]]:
    """
    Perform time series analysis on the input data.

    Parameters
    ----------

    cl_ob : ClusteringObject2D

    tau_w : int
        The time resolution for the analysis.

    t_smooth : int
        The width of the moving average for the analysis.

    Returns
    -------

    num_states : int
        Number of identified states.

    fraction_0 : float
        Fraction of unclassified data points. Between 0 and 1.

    list_of_pop : List[float]
        List of the populations of the different states.
    """
    print("* New analysis: ", tau_w, t_smooth)
    name = str(t_smooth) + "_" + str(tau_w) + "_"

    tmp_cl_ob = copy.deepcopy(cl_ob)
    tmp_cl_ob.par.tau_w = tau_w
    tmp_cl_ob.par.t_smooth = t_smooth

    tmp_cl_ob.preparing_the_data()
    tmp_cl_ob.plot_input_data(name + "Fig0")

    tmp_cl_ob, _ = iterative_search(tmp_cl_ob, name, full_out)

    if len(tmp_cl_ob.states) == 0:
        print("* No possible classification was found. ")
        # We need to free the memory otherwise it accumulates
        del tmp_cl_ob
        return 0, 1.0, [1.0]

    list_of_pop = [state.perc for state in tmp_cl_ob.states]
    fraction_0 = 1 - np.sum(list_of_pop)
    list_of_pop.insert(0, fraction_0)
    n_states = len(tmp_cl_ob.states)

    # We need to free the memory otherwise it accumulates
    del tmp_cl_ob

    print(f"Number of states identified: {n_states}, [{fraction_0}]\n")
    return n_states, fraction_0, list_of_pop


def full_output_analysis(
    cl_ob: ClusteringObject2D,
    full_out: bool,
) -> ClusteringObject2D:
    """
    Perform a comprehensive analysis on the input data.

    Parameters
    ----------

    cl_ob : ClusteringObject2D

    full_out : bool
        If True, plot all the intermediate histograms with the best fit.
        Useful for debugging.

    Returns
    -------

    cl_ob : ClusteringObject1D
        Updated with the clustering results.
    """
    cl_ob.preparing_the_data()

    cl_ob, _ = iterative_search(cl_ob, "", full_out)
    if len(cl_ob.states) == 0:
        print("* No possible classification was found. ")
        return cl_ob

    return cl_ob


def time_resolution_analysis(cl_ob: ClusteringObject2D, full_out: bool):
    """
    Performs Temporal Resolution Analysis (TRA) to explore parameter
    space and analyze the dataset.

    Args:
    - cl_ob (ClusteringObject): Conteining now only the raw input data.
    """
    tau_window_list, t_smooth_list = param_grid(
        cl_ob.par, cl_ob.data.num_of_steps
    )
    cl_ob.tau_window_list = np.array(tau_window_list)
    cl_ob.t_smooth_list = np.array(t_smooth_list)

    number_of_states = []
    fraction_0 = []
    list_of_pop: List[List[List[float]]] = [
        [[] for _ in tau_window_list] for _ in t_smooth_list
    ]

    for i, tau_w in enumerate(tau_window_list):
        tmp = [tau_w]
        tmp1 = [tau_w]
        for j, t_s in enumerate(t_smooth_list):
            n_s, f_0, l_pop = timeseries_analysis(cl_ob, tau_w, t_s, full_out)
            tmp.append(n_s)
            tmp1.append(f_0)
            list_of_pop[j][i] = l_pop
        number_of_states.append(tmp)
        fraction_0.append(tmp1)
    number_of_states_arr = np.array(number_of_states)
    fraction_0_arr = np.array(fraction_0)

    np.savetxt(
        "number_of_states.txt",
        number_of_states,
        fmt="%i",
        delimiter="\t",
        header="tau_window\t number_of_states for different t_smooth",
    )
    np.savetxt(
        "fraction_0.txt",
        fraction_0,
        delimiter=" ",
        header="tau_window\t fraction in ENV0 for different t_smooth",
    )

    cl_ob.number_of_states = number_of_states_arr
    cl_ob.fraction_0 = fraction_0_arr
    cl_ob.list_of_pop = list_of_pop


def main(
    full_output: bool = True,
    number_of_sigmas: float = 2.0,
) -> ClusteringObject2D:
    """
    Returns the clustering object with the analysi.

    all_the_input_stuff() reads the data and the parameters
    time_resolution_analysis() explore the parameter
        (tau_window, t_smooth) space.
    full_output_analysis() performs a detailed analysis
        with the chosen parameters.
    """
    print("##############################################################")
    print("# If you publish results using onion-clustering, please cite #")
    print("# this work: https://doi.org/10.48550/arXiv.2402.07786.      #")
    print("##############################################################")

    clustering_object = all_the_input_stuff(number_of_sigmas)
    time_resolution_analysis(clustering_object, full_output)
    clustering_object = full_output_analysis(clustering_object, full_output)

    return clustering_object


if __name__ == "__main__":
    main()



import os

import pytest
from onion_clustering import main as onion

### Set all the analysis parameters ###
FILE = "water_coex_100ps_1nm_LENS.npy"
PATH_TO_INPUT_DATA = "/Users/mattebecchi/00_signal_analysis/data/" + FILE
TAU_WINDOW = 10
T_DELAY = 1
NUM_TAU_W = 2
MAX_TAU_W = 10
MAX_T_SMOOTH = 2


# Define a fixture to set up the test environment
@pytest.fixture
def setup_test_environment(tmpdir):
    # tmpdir is a built-in pytest fixture providing a temporary directory
    original_dir = os.getcwd()  # Save the current working directory
    os.chdir(str(tmpdir))  # Change to the temporary directory
    yield tmpdir
    os.chdir(
        original_dir
    )  # Restore the original working directory after the test


# Define the actual test
def test_output_files(setup_test_environment):
    ### Create the 'data_directory.txt' file ###
    with open("data_directory.txt", "w+", encoding="utf-8") as file:
        print(PATH_TO_INPUT_DATA, file=file)

    ### Create the 'input_parameter.txt' file ###
    with open("input_parameters.txt", "w+", encoding="utf-8") as file:
        print(f"tau_window\t{TAU_WINDOW}", file=file)
        print(f"t_delay\t{T_DELAY}", file=file)
        print(f"num_tau_w\t{NUM_TAU_W}", file=file)
        print(f"max_tau_w\t{MAX_TAU_W}", file=file)
        print(f"max_t_smooth\t{MAX_T_SMOOTH}", file=file)

    # Call your code to generate the output files
    tmp = onion.main(False)

    # Test the output
    tmp.plot_tra_figure()
    tmp.plot_input_data("Fig0")
    tmp.plot_cumulative_figure()
    tmp.plot_one_trajectory()
    tmp.data.plot_medoids()
    tmp.plot_state_populations()
    tmp.sankey([0, 10, 20, 30, 40])
    tmp.print_labels()
    tmp.plot_pop_fractions()

    # Define the paths to the expected and actual output files
    original_dir = (
        "/Users/mattebecchi/00_signal_analysis/timeseries_analysis/test/"
    )
    expected_output_path_1 = original_dir + "output_uni/final_states.txt"
    expected_output_path_2 = original_dir + "output_uni/number_of_states.txt"
    expected_output_path_3 = original_dir + "output_uni/fraction_0.txt"
    actual_output_path_1 = "final_states.txt"
    actual_output_path_2 = "number_of_states.txt"
    actual_output_path_3 = "fraction_0.txt"

    # Use filecmp to compare the contents of the expected
    # and actual output directories
    with (
        open(expected_output_path_1, "r") as expected_file,
        open(actual_output_path_1, "r") as actual_file,
    ):
        assert expected_file.read() == actual_file.read()
    with (
        open(expected_output_path_2, "r") as expected_file,
        open(actual_output_path_2, "r") as actual_file,
    ):
        assert expected_file.read() == actual_file.read()
    with (
        open(expected_output_path_3, "r") as expected_file,
        open(actual_output_path_3, "r") as actual_file,
    ):
        assert expected_file.read() == actual_file.read()



"""
Contains the classes used for storing the clustering data.
"""

import os
from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
from matplotlib.colors import rgb2hex
from matplotlib.patches import Ellipse
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.mplot3d import Axes3D
from scipy.stats import gaussian_kde

from onion_clustering.first_classes import (
    MultiData,
    Parameters,
    StateMulti,
    StateUni,
    UniData,
)
from onion_clustering.functions import (
    gaussian,
)

COLORMAP = "viridis"


class ClusteringObject:
    """This class contains input, output and methods for plotting."""

    def __init__(
        self,
        par: Parameters,
        data: Union[UniData, MultiData],
        number_of_sigmas: float,
    ):
        self.par = par
        self.number_of_sigmas = number_of_sigmas
        self.data = data
        self.iterations = -1
        self.tau_window_list: np.ndarray
        self.t_smooth_list: np.ndarray
        self.number_of_states: np.ndarray
        self.fraction_0: np.ndarray
        self.list_of_pop: List[List[List[float]]]

    def plot_input_data(self, filename: str):
        """Plots input data for visualization."""
        raise NotImplementedError

    def preparing_the_data(self):
        """Processes raw data for analysis."""
        raise NotImplementedError

    def plot_state_populations(self):
        """Plots the populations of states over time."""

        print("* Printing populations vs time...")
        num_part = self.data.labels.shape[0]

        unique_labels = np.unique(self.data.labels)
        # If there are no assigned window, we still need the "0" state
        # for consistency:
        if 0 not in unique_labels:
            unique_labels = np.insert(unique_labels, 0, 0)

        list_of_populations = []
        for label in unique_labels:
            population = np.sum(self.data.labels == label, axis=0)
            list_of_populations.append(population / num_part)

        # Generate the color palette.
        palette = []
        n_states = unique_labels.size
        cmap = plt.get_cmap(COLORMAP, n_states)
        for i in range(cmap.N):
            rgba = cmap(i)
            palette.append(rgb2hex(rgba))

        fig, axes = plt.subplots()
        t_steps = self.data.labels.shape[1]
        time = self.par.print_time(t_steps)
        for label, pop in enumerate(list_of_populations):
            # pop_full = np.repeat(pop, self.par.tau_w)
            axes.plot(
                time, pop, label="ENV" + str(label), color=palette[label]
            )
        axes.set_xlabel(r"Time " + self.par.t_units)
        axes.set_ylabel(r"Population")
        axes.legend()

        fig.savefig("output_figures/Fig5.png", dpi=600)

    def sankey(self, tmp_frame_list: list[int]):
        """
        Computes and plots a Sankey diagram at the desired frames.

        Args:
        - tmp_frame_list (list[int]): List of frame indices.
        """
        print("* Computing and plotting the Sankey diagram...")

        all_the_labels = self.data.labels
        frame_list = np.array(tmp_frame_list)
        unique_labels = np.unique(all_the_labels)
        # If there are no assigned window, we still need the "0" state
        # for consistency:
        if 0 not in unique_labels:
            unique_labels = np.insert(unique_labels, 0, 0)
        n_states = unique_labels.size

        # Create arrays to store the source, target, and value data
        # for the Sankey diagram.
        source = np.empty((frame_list.size - 1) * n_states**2)
        target = np.empty((frame_list.size - 1) * n_states**2)
        value = np.empty((frame_list.size - 1) * n_states**2)

        # Initialize a counter variable.
        count = 0

        tmp_label1 = []
        tmp_label2 = []

        # Loop through the frame_list and calculate the transition matrix
        # for each time window.
        for i, t_0 in enumerate(frame_list[:-1]):
            # Calculate the time jump for the current time window.
            t_jump = frame_list[i + 1] - frame_list[i]

            trans_mat = np.zeros((n_states, n_states))

            # Iterate through the current time window and increment
            # the transition counts in trans_mat
            for label in all_the_labels:
                trans_mat[label[t_0]][label[t_0 + t_jump]] += 1

            # Store the source, target, and value for the Sankey diagram
            # based on trans_mat
            for j, row in enumerate(trans_mat):
                for k, elem in enumerate(row):
                    source[count] = j + i * n_states
                    target[count] = k + (i + 1) * n_states
                    value[count] = elem
                    count += 1

            # Calculate the starting and ending fractions for each state
            # and store node labels
            for j in range(n_states):
                start_fr = np.sum(trans_mat[j]) / np.sum(trans_mat)
                end_fr = np.sum(trans_mat.T[j]) / np.sum(trans_mat)
                if i == 0:
                    tmp_label1.append(f"State {j}: {start_fr * 100:.2f}%")
                tmp_label2.append(f"State {j}: {end_fr * 100:.2f}%")

        arr_label1 = np.array(tmp_label1)
        arr_label2 = np.array(tmp_label2).flatten()

        # Concatenate the temporary labels to create the final node labels.
        label = np.concatenate((arr_label1, arr_label2))

        # Generate a color palette for the Sankey diagram.
        palette = []
        cmap = plt.get_cmap(COLORMAP, n_states)
        for i in range(cmap.N):
            rgba = cmap(i)
            palette.append(rgb2hex(rgba))

        # Tile the color palette to match the number of frames.
        color = np.tile(palette, frame_list.size)

        # Create dictionaries to define the Sankey diagram nodes and links.
        node = {"label": label, "pad": 30, "thickness": 20, "color": color}
        link = {"source": source, "target": target, "value": value}

        # Create the Sankey diagram using Plotly.
        sankey_data = go.Sankey(
            link=link, node=node, arrangement="perpendicular"
        )
        fig = go.Figure(sankey_data)

        # Add the title with the time information.
        fig.update_layout(
            title="Frames: "
            + str(frame_list * self.par.t_conv)
            + " "
            + self.par.t_units
        )

        fig.write_image("output_figures/Fig6.png", scale=5.0)

    def create_all_the_labels(self) -> np.ndarray:
        """
        Assigns labels to individual frames by repeating the existing labels.

        Returns:
        - np.ndarray: An updated ndarray with labels assigned
            to individual frames by repeating the existing labels.
        """
        all_the_labels = np.repeat(self.data.labels, self.par.tau_w, axis=1)
        return all_the_labels

    def plot_cumulative_figure(self):
        """Plots clustering output with Gaussians and threshols."""
        raise NotImplementedError

    def plot_one_trajectory(self):
        """Plots the colored trajectory of an example particle."""
        raise NotImplementedError

    def print_colored_trj_from_xyz(self, trj_file: str):
        """
        Creates a new XYZ file ('colored_trj.xyz') by coloring
        the original trajectory based on cluster labels.

        Args:
        - trj_file (str): Path to the original XYZ trajectory file.
        """
        if os.path.exists(trj_file):
            print("* Loading trajectory.xyz...")
            with open(trj_file, "r", encoding="utf-8") as in_file:
                tmp = [line.strip().split() for line in in_file]

            all_the_labels = self.create_all_the_labels()
            num_of_particles = all_the_labels.shape[0]
            total_time = all_the_labels.shape[1]
            nlines = (num_of_particles + 2) * total_time

            frames_to_remove = int(self.par.t_smooth / 2) + self.par.t_delay
            print("\t Removing the first", frames_to_remove, "frames...")
            tmp = tmp[frames_to_remove * (num_of_particles + 2) :]

            frames_to_remove = int(
                (len(tmp) - nlines) / (num_of_particles + 2)
            )
            print("\t Removing the last", frames_to_remove, "frames...")
            tmp = tmp[:nlines]

            with open("colored_trj.xyz", "w+", encoding="utf-8") as out_file:
                i = 0
                for j in range(total_time):
                    print(tmp[i][0], file=out_file)
                    print("Properties=species:S:1:pos:R:3", file=out_file)
                    for k in range(num_of_particles):
                        print(
                            all_the_labels[k][j],
                            tmp[i + 2 + k][1],
                            tmp[i + 2 + k][2],
                            tmp[i + 2 + k][3],
                            file=out_file,
                        )
                    i += num_of_particles + 2
        else:
            print(f"No {trj_file} found for coloring the trajectory.")

    def print_labels(self):
        """
        Print the label for every particle for every frame.
        Output is a (N, T) array in a .npy file.
        """
        print("* Print labels for all the data points...")
        all_the_labels = self.create_all_the_labels()
        np.save("all_labels.npy", all_the_labels)

    def print_signal_with_labels(self):
        """
        Creates a file ('signal_with_labels.dat') with signal values
        and associated cluster labels.
        """
        m_clean = self.data.matrix
        all_the_labels = self.create_all_the_labels()

        with open("signal_with_labels.dat", "w+", encoding="utf-8") as file:
            if m_clean.shape[2] == 2:
                print("Signal 1 Signal 2 Cluster Frame", file=file)
            else:
                print("Signal 1 Signal 2 Signal 3 Cluster Frame", file=file)
            for j in range(all_the_labels.shape[1]):
                for i in range(all_the_labels.shape[0]):
                    if m_clean.shape[2] == 2:
                        print(
                            m_clean[i][j][0],
                            m_clean[i][j][1],
                            all_the_labels[i][j],
                            j + 1,
                            file=file,
                        )
                    else:
                        print(
                            m_clean[i][j][0],
                            m_clean[i][j][1],
                            m_clean[i][j][2],
                            all_the_labels[i][j],
                            j + 1,
                            file=file,
                        )

    def plot_tra_figure(self):
        """Plots time resolution analysis figures."""

        t_conv, units = self.par.t_conv, self.par.t_units
        min_t_s, max_t_s = self.par.min_t_smooth, self.par.max_t_smooth
        step_t_s = self.par.step_t_smooth

        time = self.number_of_states.T[0] * t_conv
        number_of_states = self.number_of_states[:, 1:].T
        fraction_0 = self.fraction_0[:, 1:].T

        for i, t_smooth in enumerate(range(min_t_s, max_t_s + 1, step_t_s)):
            fig, axes = plt.subplots()
            y_signal = number_of_states[i]
            y_2 = fraction_0[i]

            ### General plot settings ###
            axes.plot(time, y_signal, marker="o")
            axes.set_xlabel(r"Time resolution $\Delta t$ " + units)
            axes.set_ylabel(r"# environments", weight="bold", c="#1f77b4")
            axes.set_xscale("log")
            axes.set_xlim(time[0] * 0.75, time[-1] * 1.5)
            axes.yaxis.set_major_locator(MaxNLocator(integer=True))

            ### Top x-axes settings ###
            axes2 = axes.twiny()
            axes2.set_xlabel(r"Time resolution $\Delta t$ [frames]")
            axes2.set_xscale("log")
            axes2.set_xlim(time[0] * 0.75 / t_conv, time[-1] * 1.5 / t_conv)

            axesr = axes.twinx()
            axesr.plot(time, y_2, marker="o", c="#ff7f0e")
            axesr.set_ylabel("Population of env 0", weight="bold", c="#ff7f0e")

            fig.savefig(
                "output_figures/Time_resolution_analysis_"
                + str(t_smooth)
                + ".png",
                dpi=600,
            )

    def plot_pop_fractions(self):
        """
        Plot, for every time resolution, the populations of the ENVs.

        The bottom state is the ENV0.
        """
        print("* Print populations fractions...")

        t_conv, units = self.par.t_conv, self.par.t_units
        time = self.tau_window_list * t_conv

        for i, t_smooth in enumerate(self.t_smooth_list):
            pop_array = self.list_of_pop[i]

            max_num_of_states = np.max(
                [len(pop_list) for pop_list in pop_array]
            )
            for _, pop_list in enumerate(pop_array):
                while len(pop_list) < max_num_of_states:
                    pop_list.append(0.0)

            pop_array = np.array(pop_array)

            fig, axes = plt.subplots()
            width = time[1:] - time[:-1]
            width = np.insert(width, 0, width[0] / 2)

            bottom = np.zeros(len(pop_array))
            for _, state in enumerate(pop_array.T):
                _ = axes.bar(
                    time, state, width, bottom=bottom, edgecolor="black"
                )
                bottom += state

            axes.set_xlabel(rf"Time resolution $\Delta t$ {units}")
            axes.set_ylabel(r"Population's fractions")
            axes.set_xscale("log")

            fig.savefig(
                f"output_figures/Populations_{t_smooth}.png",
                dpi=600,
            )


class ClusteringObject1D(ClusteringObject):
    """This class contains input, output and methods for plotting."""

    states: List[StateUni] = []

    def plot_input_data(self, filename: str):
        """Plots input data for visualization."""

        # Flatten the m_clean matrix and compute histogram counts and bins
        m_clean = self.data.matrix
        flat_m = m_clean.flatten()

        kde = gaussian_kde(flat_m)
        if self.par.bins == "auto":
            n_bins = 100
        else:
            n_bins = int(self.par.bins)
        bins = np.linspace(np.min(flat_m), np.max(flat_m), n_bins)
        counts = kde.evaluate(bins)

        bins -= (bins[1] - bins[0]) / 2
        counts *= flat_m.size

        # Create a plot with two subplots (side-by-side)
        fig, axes = plt.subplots(
            1,
            2,
            sharey=True,
            gridspec_kw={"width_ratios": [3, 1]},
            figsize=(9, 4.8),
        )

        # Plot histogram in the second subplot (right side)
        axes[1].stairs(counts[:-1], bins, fill=True, orientation="horizontal")

        # Plot the individual trajectories in the first subplot (left side)
        time = self.par.print_time(m_clean.shape[1])
        step = 10 if m_clean.size > 1000000 else 1
        for mol in m_clean[::step]:
            axes[0].plot(
                time, mol, c="xkcd:black", lw=0.1, alpha=0.5, rasterized=True
            )

        # Set labels and titles for the plots
        axes[0].set_ylabel("Signal")
        axes[0].set_xlabel(rf"Simulation time $t$ {self.par.t_units}")
        axes[1].set_xticklabels([])

        fig.savefig("output_figures/Fig0.png", dpi=600)
        plt.close(fig)

    def preparing_the_data(self):
        """Processes raw data for analysis."""

        tau_window, t_smooth = self.par.tau_w, self.par.t_smooth
        t_conv, t_units = self.par.t_conv, self.par.t_units

        # Apply filtering on the data
        self.data.smooth_mov_av(t_smooth)  # Smoothing using moving average

        # Normalize the data to the range [0, 1]. Usually not needed. ###
        # self.data.normalize()

        # Calculate the number of windows for the analysis.
        num_windows = int(self.data.num_of_steps / tau_window)

        # Print informative messages about trajectory details.
        print(
            "\tTrajectory has "
            + str(self.data.num_of_particles)
            + " particles. "
        )
        print(
            "\tTrajectory of length "
            + str(self.data.num_of_steps)
            + " frames ("
            + str(self.data.num_of_steps * t_conv),
            t_units + ")",
        )
        print(
            "\tUsing "
            + str(num_windows)
            + " windows of length "
            + str(tau_window)
            + " frames ("
            + str(tau_window * t_conv),
            t_units + ")",
        )

    def plot_cumulative_figure(self):
        """Plots clustering output with Gaussians and threshols."""

        print("* Printing cumulative figure...")

        # Compute histogram of flattened self.data.matrix
        flat_m = self.data.matrix.flatten()

        kde = gaussian_kde(flat_m)
        if self.par.bins == "auto":
            n_bins = 100
        else:
            n_bins = int(self.par.bins)
        bins = np.linspace(np.min(flat_m), np.max(flat_m), n_bins)
        counts = kde.evaluate(bins)

        # counts, bins = np.histogram(flat_m, bins=self.par.bins, density=True)
        bins -= (bins[1] - bins[0]) / 2
        counts *= flat_m.size

        # gap = 1
        # if bins.size > 49:
        #     gap = int(bins.size * 0.02) * 2
        # counts = moving_average(counts, gap)
        # bins = moving_average(bins, gap)

        # Create a 1x2 subplots with shared y-axis
        fig, axes = plt.subplots(
            1,
            2,
            sharey=True,
            gridspec_kw={"width_ratios": [3, 1]},
            figsize=(9, 4.8),
        )

        # Plot the histogram on the right subplot (axes[1])
        axes[1].stairs(
            counts[:-1], bins, fill=True, orientation="horizontal", alpha=0.5
        )

        # Create a color palette for plotting states
        palette = []
        n_states = len(self.states)
        cmap = plt.get_cmap(COLORMAP, n_states + 1)
        for i in range(1, cmap.N):
            rgba = cmap(i)
            palette.append(rgb2hex(rgba))

        # Define time and y-axis limits for the left subplot (axes[0])
        y_spread = np.max(self.data.matrix) - np.min(self.data.matrix)
        y_lim = [
            np.min(self.data.matrix) - 0.025 * y_spread,
            np.max(self.data.matrix) + 0.025 * y_spread,
        ]
        time = self.par.print_time(self.data.matrix.shape[1])

        # Plot the individual trajectories on the left subplot (axes[0])
        step = 10 if self.data.matrix.size > 1000000 else 1
        for mol in self.data.matrix[::step]:
            axes[0].plot(
                time,
                mol,
                c="xkcd:black",
                ms=0.1,
                lw=0.1,
                alpha=0.5,
                rasterized=True,
            )

        for state_id, state in enumerate(self.states):
            popt = [state.mean, state.sigma, state.area]
            axes[1].plot(
                gaussian(np.linspace(bins[0], bins[-1], 1000), *popt),
                np.linspace(bins[0], bins[-1], 1000),
                color=palette[state_id],
            )

        style_color_map = {
            0: ("--", "xkcd:black"),
            1: ("--", "xkcd:blue"),
            2: ("--", "xkcd:red"),
        }

        time2 = np.linspace(
            time[0] - 0.05 * (time[-1] - time[0]),
            time[-1] + 0.05 * (time[-1] - time[0]),
            100,
        )
        for state_id, state in enumerate(self.states):
            linestyle, color = style_color_map.get(
                state.th_inf[1], ("-", "xkcd:black")
            )
            axes[1].hlines(
                state.th_inf[0],
                xmin=0.0,
                xmax=np.amax(counts),
                linestyle=linestyle,
                color=color,
            )
            axes[0].fill_between(
                time2,
                state.th_inf[0],
                state.th_sup[0],
                color=palette[state_id],
                alpha=0.25,
            )
        if len(self.states) > 0:
            axes[1].hlines(
                self.states[-1].th_sup[0],
                xmin=0.0,
                xmax=np.amax(counts),
                linestyle=linestyle,
                color="black",
            )

        # Set plot titles and axis labels
        axes[0].set_ylabel("Signal")
        axes[0].set_xlabel(r"Time $t$ " + self.par.t_units)
        axes[0].set_xlim([time2[0], time2[-1]])
        axes[0].set_ylim(y_lim)
        axes[1].set_xticklabels([])

        fig.savefig("output_figures/Fig2.png", dpi=600)
        plt.close(fig)

    def plot_one_trajectory(self):
        """Plots the colored trajectory of an example particle."""

        example_id = self.par.example_id
        # Get the signal of the example particle
        all_the_labels = self.create_all_the_labels()
        signal = self.data.matrix[example_id][: all_the_labels.shape[1]]

        # Create time values for the x-axis
        time = self.par.print_time(all_the_labels.shape[1])

        # Create a figure and axes for the plot
        fig, axes = plt.subplots()

        unique_labels = np.unique(all_the_labels)
        # If there are no assigned window, we still need the "0" state
        # for consistency:
        if 0 not in unique_labels:
            unique_labels = np.insert(unique_labels, 0, 0)

        cmap = plt.get_cmap(
            COLORMAP, np.max(unique_labels) - np.min(unique_labels) + 1
        )
        color = all_the_labels[example_id]
        axes.plot(time, signal, c="black", lw=0.1)

        axes.scatter(
            time,
            signal,
            c=color,
            cmap=cmap,
            vmin=np.min(unique_labels),
            vmax=np.max(unique_labels),
            s=1.0,
        )

        # Add title and labels to the axes
        fig.suptitle("Example particle: ID = " + str(example_id))
        axes.set_xlabel("Time " + self.par.t_units)
        axes.set_ylabel("Signal")

        fig.savefig("output_figures/Fig3.png", dpi=600)
        plt.close(fig)


class ClusteringObject2D(ClusteringObject):
    """This class contains input, output and methods for plotting."""

    states: List[StateMulti] = []

    def plot_input_data(self, filename: str):
        """
        Plot input data: histograms and trajectories.

        This function creates plots for input data:
        - For 2D data: Creates histograms and individual trajectories
            (side-by-side).
        - For 3D data: Creates a 3D plot showing the trajectories.
        """
        bin_selection = []
        counts_selection = []
        m_clean = self.data.matrix

        if isinstance(self.data, MultiData):
            for dim in range(self.data.dims):
                # Flatten the m matrix and compute histogram counts and bins
                flat_m = m_clean[:, :, dim].flatten()
                counts0, bins0 = np.histogram(
                    flat_m, bins=self.par.bins, density=True
                )
                counts0 *= flat_m.size
                bin_selection.append(bins0)
                counts_selection.append(counts0)

            if self.data.dims == 2:
                # Create a plot with two subplots (side-by-side)
                fig = plt.figure(figsize=(9, 9))
                grid = fig.add_gridspec(4, 4)
                ax1 = fig.add_subplot(grid[0:1, 0:3])
                ax2 = fig.add_subplot(grid[1:4, 0:3])
                ax3 = fig.add_subplot(grid[1:4, 3:4])
                ax1.set_xticklabels([])
                ax3.set_yticklabels([])

                # Plot histograms
                ax1.stairs(counts_selection[0], bin_selection[0], fill=True)
                ax3.stairs(
                    counts_selection[1],
                    bin_selection[1],
                    fill=True,
                    orientation="horizontal",
                )

                # Plot the individual trajectories in the first subplot
                id_max, id_min = 0, 0
                for idx, mol in enumerate(m_clean):
                    if np.max(mol) == np.max(m_clean):
                        id_max = idx
                    if np.min(mol) == np.min(m_clean):
                        id_min = idx
                step = 10 if m_clean.size > 1000000 else 1
                for idx, mol in enumerate(m_clean[::step]):
                    ax2.plot(
                        mol[:, 0],
                        mol[:, 1],
                        color="black",
                        lw=0.1,
                        alpha=0.5,
                        rasterized=True,
                    )
                ax2.plot(
                    m_clean[id_min][:, 0],
                    m_clean[id_min][:, 1],
                    color="black",
                    lw=0.1,
                    alpha=0.5,
                    rasterized=True,
                )
                ax2.plot(
                    m_clean[id_max][:, 0],
                    m_clean[id_max][:, 1],
                    color="black",
                    lw=0.1,
                    alpha=0.5,
                    rasterized=True,
                )

                # Set labels and titles for the plots
                ax2.set_ylabel("Signal 1")
                ax2.set_xlabel("Signal 2")

            elif self.data.dims == 3:
                fig = plt.figure(figsize=(6, 6))
                axes: Axes3D = fig.add_subplot(111, projection="3d")

                # Plot the individual trajectories
                step = 1 if m_clean.size > 1000000 else 1
                for idx, mol in enumerate(m_clean[::step]):
                    axes.plot(
                        mol[:, 0],
                        mol[:, 1],
                        mol[:, 2],
                        color="black",
                        marker="o",
                        ms=0.5,
                        lw=0.2,
                        alpha=1.0,
                        rasterized=True,
                    )

                # Set labels and titles for the plots
                axes.set_xlabel("Signal 1")
                axes.set_ylabel("Signal 2")
                axes.set_zlabel("Signal 3")

            fig.savefig("output_figures/" + filename + ".png", dpi=600)
            plt.close(fig)

    def preparing_the_data(self):
        """
        Prepare the raw data for analysis.

        This function prepares the raw data for analysis:
        - Applies a moving average filter on the raw data.
        - Normalizes the data to the range [0, 1] (commented out in the code).
        - Calculates the number of windows for analysis based on parameters.
        - Prints informative messages about trajectory details.
        """
        tau_window, t_smooth = self.par.tau_w, self.par.t_smooth
        t_conv, t_units = self.par.t_conv, self.par.t_units

        self.data.smooth(t_smooth)
        ### Normalizes data in [0, 1]. Usually not necessary.
        ### The arg is the list of components to NOT normalize
        # self.data.normalize([])

        # Calculate the number of windows for the analysis.
        num_windows = int(self.data.num_of_steps / tau_window)

        # Print informative messages about trajectory details.
        print(
            "\tTrajectory has "
            + str(self.data.num_of_particles)
            + " particles. "
        )
        print(
            "\tTrajectory of length "
            + str(self.data.num_of_steps)
            + " frames ("
            + str(self.data.num_of_steps * t_conv)
            + " "
            + t_units
            + ")."
        )
        print(
            "\tUsing "
            + str(num_windows)
            + " windows of length "
            + str(tau_window)
            + " frames ("
            + str(tau_window * t_conv)
            + " "
            + t_units
            + ")."
        )

    def plot_cumulative_figure(self):
        """
        Plot a cumulative figure showing trajectories and identified states.
        """
        print("* Printing cumulative figure...")
        n_states = len(self.states) + 1
        tmp = plt.get_cmap(COLORMAP, n_states)
        colors_from_cmap = tmp(np.arange(0, 1, 1 / n_states))
        colors_from_cmap[-1] = tmp(1.0)
        m_clean = self.data.matrix
        all_the_labels = self.create_all_the_labels()

        if m_clean.shape[2] == 3:
            fig, axes = plt.subplots(2, 2, figsize=(6, 6))
            dir0 = [0, 0, 1]
            dir1 = [1, 2, 2]
            ax0 = [0, 0, 1]
            ax1 = [0, 1, 0]

            for k in range(3):
                d_0 = dir0[k]
                d_1 = dir1[k]
                a_0 = ax0[k]
                a_1 = ax1[k]
                # Plot the individual trajectories
                id_max, id_min = 0, 0
                for idx, mol in enumerate(m_clean):
                    if np.max(mol) == np.max(m_clean):
                        id_max = idx
                    if np.min(mol) == np.min(m_clean):
                        id_min = idx

                line_w = 0.05
                max_t = all_the_labels.shape[1]
                m_resized = m_clean[:, :max_t:, :]
                step = 5 if m_resized.size > 1000000 else 1

                for i, mol in enumerate(m_resized[::step]):
                    axes[a_0][a_1].plot(
                        mol.T[d_0],
                        mol.T[d_1],
                        c="black",
                        lw=line_w,
                        rasterized=True,
                        zorder=0,
                    )
                    color_list = all_the_labels[i * step]
                    axes[a_0][a_1].scatter(
                        mol.T[d_0],
                        mol.T[d_1],
                        c=color_list,
                        cmap=COLORMAP,
                        vmin=0,
                        vmax=n_states - 1,
                        s=0.5,
                        rasterized=True,
                    )

                    color_list = all_the_labels[id_min]
                    axes[a_0][a_1].plot(
                        m_resized[id_min].T[d_0],
                        m_resized[id_min].T[d_1],
                        c="black",
                        lw=line_w,
                        rasterized=True,
                        zorder=0,
                    )
                    axes[a_0][a_1].scatter(
                        m_resized[id_min].T[d_0],
                        m_resized[id_min].T[d_1],
                        c=color_list,
                        cmap=COLORMAP,
                        vmin=0,
                        vmax=n_states - 1,
                        s=0.5,
                        rasterized=True,
                    )
                    color_list = all_the_labels[id_max]
                    axes[a_0][a_1].plot(
                        m_resized[id_max].T[d_0],
                        m_resized[id_max].T[d_1],
                        c="black",
                        lw=line_w,
                        rasterized=True,
                        zorder=0,
                    )
                    axes[a_0][a_1].scatter(
                        m_resized[id_max].T[d_0],
                        m_resized[id_max].T[d_1],
                        c=color_list,
                        cmap=COLORMAP,
                        vmin=0,
                        vmax=n_states - 1,
                        s=0.5,
                        rasterized=True,
                    )

                    # Plot the Gaussian distributions of states
                    if k == 0:
                        for state in self.states:
                            ellipse = Ellipse(
                                tuple(state.mean),
                                state.axis[d_0],
                                state.axis[d_1],
                                color="black",
                                fill=False,
                            )
                            axes[a_0][a_1].add_patch(ellipse)

                # Set plot titles and axis labels
                axes[a_0][a_1].set_xlabel(r"Signal " + str(d_0))
                axes[a_0][a_1].set_ylabel(r"Signal " + str(d_1))

            axes[1][1].axis("off")
            fig.savefig("output_figures/Fig2.png", dpi=600)
            plt.close(fig)

        elif m_clean.shape[2] == 2:
            fig, axes = plt.subplots(figsize=(6, 6))

            # Plot the individual trajectories
            id_max, id_min = 0, 0
            for idx, mol in enumerate(m_clean):
                if np.max(mol) == np.max(m_clean):
                    id_max = idx
                if np.min(mol) == np.min(m_clean):
                    id_min = idx

            line_w = 0.05
            max_t = all_the_labels.shape[1]
            m_resized = m_clean[:, :max_t:, :]
            step = 5 if m_resized.size > 1000000 else 1

            for i, mol in enumerate(m_resized[::step]):
                axes.plot(
                    mol.T[0],
                    mol.T[1],
                    c="black",
                    lw=line_w,
                    rasterized=True,
                    zorder=0,
                )
                color_list = all_the_labels[i * step]
                axes.scatter(
                    mol.T[0],
                    mol.T[1],
                    c=color_list,
                    cmap=COLORMAP,
                    vmin=0,
                    vmax=n_states - 1,
                    s=0.5,
                    rasterized=True,
                )

            color_list = all_the_labels[id_min]
            axes.plot(
                m_resized[id_min].T[0],
                m_resized[id_min].T[1],
                c="black",
                lw=line_w,
                rasterized=True,
                zorder=0,
            )
            axes.scatter(
                m_resized[id_min].T[0],
                m_resized[id_min].T[1],
                c=color_list,
                cmap=COLORMAP,
                vmin=0,
                vmax=n_states - 1,
                s=0.5,
                rasterized=True,
            )
            color_list = all_the_labels[id_max]
            axes.plot(
                m_resized[id_max].T[0],
                m_resized[id_max].T[1],
                c="black",
                lw=line_w,
                rasterized=True,
                zorder=0,
            )
            axes.scatter(
                m_resized[id_max].T[0],
                m_resized[id_max].T[1],
                c=color_list,
                cmap=COLORMAP,
                vmin=0,
                vmax=n_states - 1,
                s=0.5,
                rasterized=True,
            )

            # Plot the Gaussian distributions of states
            for state in self.states:
                ellipse = Ellipse(
                    tuple(state.mean),
                    state.axis[0],
                    state.axis[1],
                    color="black",
                    fill=False,
                )
                axes.add_patch(ellipse)

            # Set plot titles and axis labels
            axes.set_xlabel(r"$x$")
            axes.set_ylabel(r"$y$")

            fig.savefig("output_figures/Fig2.png", dpi=600)
            plt.close(fig)

    def plot_one_trajectory(self):
        """Plots the colored trajectory of an example particle."""

        m_clean = self.data.matrix
        all_the_labels = self.create_all_the_labels()

        # Get the signal of the example particle
        signal_x = m_clean[self.par.example_id].T[0][: all_the_labels.shape[1]]
        signal_y = m_clean[self.par.example_id].T[1][: all_the_labels.shape[1]]

        fig, axes = plt.subplots(figsize=(6, 6))

        # Create a colormap to map colors to the labels
        cmap = plt.get_cmap(
            COLORMAP,
            int(
                np.max(np.unique(all_the_labels))
                - np.min(np.unique(all_the_labels))
                + 1
            ),
        )
        color = all_the_labels[self.par.example_id]
        axes.plot(signal_x, signal_y, c="black", lw=0.1)

        axes.scatter(
            signal_x,
            signal_y,
            c=color,
            cmap=cmap,
            vmin=np.min(np.unique(all_the_labels)),
            vmax=np.max(np.unique(all_the_labels)),
            s=1.0,
            zorder=10,
        )

        # Set plot titles and axis labels
        fig.suptitle("Example particle: ID = " + str(self.par.example_id))
        axes.set_xlabel(r"$x$")
        axes.set_ylabel(r"$y$")

        fig.savefig("output_figures/Fig3.png", dpi=600)
        plt.close(fig)



import os

import pytest
from onion_clustering import main_2d


# Define a fixture to set up the test environment
@pytest.fixture
def setup_test_environment(tmpdir):
    # tmpdir is a built-in pytest fixture providing a temporary directory
    original_dir = os.getcwd()  # Save the current working directory
    os.chdir(str(tmpdir))  # Change to the temporary directory
    yield tmpdir
    os.chdir(
        original_dir
    )  # Restore the original working directory after the test


# Define the actual test
def test_output_files(setup_test_environment):
    ### Set all the analysis parameters ###
    PATH_TO_INPUT_DATA = (
        "/Users/mattebecchi/00_signal_analysis/data/"
        "synthetic_2D/3D_synthetic_data.npy"
    )
    TAU_WINDOW = 10
    T_CONV = 200
    T_UNITS = "dt"
    NUM_TAU_W = 2
    MAX_TAU_W = 10
    MAX_T_SMOOTH = 2
    BINS = 50

    ### Create the 'data_directory.txt' file ###
    with open("data_directory.txt", "w+", encoding="utf-8") as file:
        print(PATH_TO_INPUT_DATA, file=file)

    ### Create the 'input_parameter.txt' file ###
    with open("input_parameters.txt", "w+", encoding="utf-8") as file:
        print("tau_window\t" + str(TAU_WINDOW), file=file)
        print("t_conv\t" + str(T_CONV), file=file)
        print("t_units\t" + T_UNITS, file=file)
        print("num_tau_w\t" + str(NUM_TAU_W), file=file)
        print("max_tau_w\t" + str(MAX_TAU_W), file=file)
        print("max_t_smooth\t" + str(MAX_T_SMOOTH), file=file)
        print("bins\t" + str(BINS), file=file)

    # Call your code to generate the output files
    tmp = main_2d.main()

    # Test the output
    tmp.plot_tra_figure()
    tmp.plot_pop_fractions()
    tmp.plot_input_data("Fig0")
    tmp.plot_cumulative_figure()
    tmp.plot_one_trajectory()
    tmp.data.plot_medoids()
    tmp.plot_state_populations()
    tmp.sankey([0, 100, 200, 300, 400])
    tmp.print_labels()

    # Define the paths to the expected and actual output files
    original_dir = (
        "/Users/mattebecchi/00_signal_analysis/timeseries_analysis/test/"
    )
    expected_output_path_1 = original_dir + "output_multi/final_states.txt"
    expected_output_path_2 = original_dir + "output_multi/number_of_states.txt"
    expected_output_path_3 = original_dir + "output_multi/fraction_0.txt"
    actual_output_path_1 = "final_states.txt"
    actual_output_path_2 = "number_of_states.txt"
    actual_output_path_3 = "fraction_0.txt"

    # Use filecmp to compare the contents of the expected and actual
    # output directories
    with (
        open(expected_output_path_1, "r") as expected_file,
        open(actual_output_path_1, "r") as actual_file,
    ):
        assert expected_file.read() == actual_file.read()
    with (
        open(expected_output_path_2, "r") as expected_file,
        open(actual_output_path_2, "r") as actual_file,
    ):
        assert expected_file.read() == actual_file.read()
    with (
        open(expected_output_path_3, "r") as expected_file,
        open(actual_output_path_3, "r") as actual_file,
    ):
        assert expected_file.read() == actual_file.read()
